--
-- Fran static image rendering using Win32
--
-- Converting Picture values into graphical output (Win32 GDI).
--
-- Last modified Sun Nov 10 16:19:06 1996

----------------------------------------------------------------
-- ToDo:
--
-- o Use bounding boxes more pervasively.  
--   Define a function 
--  
--     easyBBox :: Image -> Maybe (Point2, Point2) 
--  
--   that picks off the easy cases:
--     Circle, Square, Bitmap, Bezier, Text, and BBoxed2.
--
----------------------------------------------------------------

module RenderImage (
	draw
	) where

import StaticTypes(
	Length, Radians,
	Vector2, vector2XY, vector2XYCoords, vector2PolarCoords,
	Point2,  point2XY, point2XYCoords, origin2,
	Transform2, compose2, inverse2, identity2, 
	Transformable2((*%)),
	Scaleable2(scale2),
	Image(..), squarePoints, circleRadius, 
	TextT(..), Family(..),
	Font(..)
	)

import GraphicsPicture hiding (Font)

----------------------------------------------------------------
-- The interface
----------------------------------------------------------------

draw :: Image -> Picture

----------------------------------------------------------------
-- The implementation
----------------------------------------------------------------

draw im = 
  -- Win32.getStockPen Win32.wHITE_PEN     >>= \ pen ->
  -- Win32.selectPen hdc pen               >>
  withBkMode Transparent               $
  withTextAlignment (Center, Baseline) $
  draw' identity2 False im

draw' :: Transform2 -> Bool -> Image -> Picture
draw' tr hasColor im =
  case im of
  EmptyImage         -> empty
  Circle 
    | easy           -> ellipse       p1 p2    -- fast and accurate
    | otherwise      -> shearEllipse  p0 p1 p2 -- more general, slower, approximate
   where
    easy = unrotated p0 p1 p2
    p0 = f circ0
    p1 = f circ1
    p2 = f circ2

  Square             -> polygon       (map f squarePoints)

  -- This works on both Win'95 and Win'NT
  Bitmap size bmp    -> stretchBitmap (f p0) (f p3) bmp
   where
    p0 = scale2 size *% bitmap0
    p3 = scale2 size *% bitmap3
{-
  -- This only works on Win'NT (but looks much better)
  Bitmap size bmp    -> shearBitmap   (f p0) (f p1) (f p2) bmp
   where
    p0 = scale2 size *% bitmap0
    p1 = scale2 size *% bitmap1
    p2 = scale2 size *% bitmap2
-}
  Line p1 p2         -> line          (f p1) (f p2)
  Polyline pts       -> polyline      (map f pts)
  Polygon pts        -> polygon       (map f pts)
  Bezier p1 p2 p3 p4 -> polyBezier    (map f [p1,p2,p3,p4])
  RenderedText str   -> textTform     (f text0) (g text1) str
  WithColor c im 
    | hasColor       -> draw' tr hasColor im
    | otherwise      -> withRGB (rgb c) (draw' tr True im)
  Over p q           -> draw' tr hasColor p `over` draw' tr hasColor q
  TransformedImage tr' p -> draw' (tr `compose2` tr') hasColor p
  BBoxed2 p1 p2 im
    | tooSmall       -> empty
    | otherwise      -> draw' tr hasColor im
   where
    tooSmall = maxPixelDist (f p1) (f p2) <= minPixelDrawSize

 where
  f :: Point2 -> Point
  f pt = case point2XYCoords (tr *% pt) of (x,y) -> (round x, round y)

  g :: Vector2 -> (Dimension, Angle)
  g v = case vector2PolarCoords (tr *% v) of (x,y) -> (round x, y)

----------------------------------------------------------------
-- Utilities
----------------------------------------------------------------

-- Test whether we can use "ellipse" to draw a transformed circle
-- or whether we're going to have to use the much slower, less
-- accurate "shearEllipse" function.

-- We test whether the points defining the circle have had a
-- "non-trivial" rotation applied to them.  We could test the
-- transformation instead - but a "general purpose" rotation
-- detection wouldn't spot the special case that rotated circles
-- are still circles.

unrotated :: Point -> Point -> Point -> Bool
unrotated (x0,y0) (x1,y1) (x2,y2) =
  -- True &&                  -- don't care, go fast no matter what
  (  (dx1+dx2) == (dy1+dy2)   -- circle
  || (dx2 == 0 && dy1 == 0)   -- unrotated ellipse
  || (dx1 == 0 && dy2 == 0)   -- ellipse rotated 90 degrees
  )
 where
  dx1 = x1 - x0
  dy1 = y1 - y0
  dx2 = x2 - x0
  dy2 = y2 - y0

-- Color conversion
rgb :: StaticTypes.Color -> RGB
rgb c = case StaticTypes.toRGB c of { StaticTypes.RGB r g b -> 
        RGB (round (255*r)) (round (255*g)) (round (255*b))
        }

-- There are two main options in choosing a fudge factor:
--
-- o If "fudge = k * length str", then the string will occupy the
--   same area on the screen no matter how long it is.  (At least,
--   it will for fixed-width fonts - proportional width fonts
--   can vary a little.)  Zero length strings present a problem!
--
-- o If "fudge = k", then each character will occupy the same 
--   area on the screen no matter how long the string is.
--
-- The second choice seems to produce a less surprising system.
--
-- (No matter which approach you choose, it seems that the
--  choice of "k" has to be empirically derived.)

-- draw some text along a line
textTform :: Point -> (Dimension, Angle) -> TextT -> Picture
textTform pt (len, angle) (TextT (Font fam bold italic) str) =
  mkFont (width,height) angle bold italic (fontName fam) $ \ font ->
  withFont font $
  text pt str
 where
  fudge  = 15
  width  = round (fromInt len / fudge)
  height = 2*width

fontName :: Family -> String
fontName System     = "System" -- this is going to break ..
fontName TimesRoman = "Times New Roman"
fontName Courier    = "Courier New"
fontName Arial      = "Arial"
fontName Symbol     = "Symbol"

maxPixelDist :: Point -> Point -> Dimension
maxPixelDist (x1,y1) (x2,y2) = abs (x1-x2) `max` abs (y1-y2)

----------------------------------------------------------------
-- Parameters
----------------------------------------------------------------

-- If the x or y pixel separation is below minPixelDrawSize, don't draw.
-- To do: cull off-window bboxed images.

minPixelDrawSize :: Dimension
minPixelDrawSize = 1

-- The location of text is defined by its centre and a vector to its 
-- right hand side.
-- The fit should (could?) be exact for fixed width fonts but 
-- will is rather approximate for variable width fonts.

text0 :: Point2
text0 = origin2

text1 :: Vector2
text1 = vector2XY 1 0

-- A transformed circle is defined by three corners of a parallelogram.

circ0, circ1, circ2 :: Point2
circ0 = point2XY (-circleRadius) (-circleRadius)
circ1 = point2XY   circleRadius  (-circleRadius)
circ2 = point2XY (-circleRadius)   circleRadius 

-- A transformed bitmap is defined by three corners of a parallelogram
-- or by opposite corners of a rectangle (if we don't rotate it).

bitmap0, bitmap1, bitmap2, bitmap3 :: Point2
bitmap0 = point2XY (-0.5) (-0.5)
bitmap1 = point2XY ( 0.5) (-0.5)
bitmap2 = point2XY (-0.5) ( 0.5)
bitmap3 = point2XY ( 0.5) ( 0.5)

----------------------------------------------------------------
-- End
----------------------------------------------------------------
