--
-- Fran animation display driver
-- 
-- Last modified Tue May 06 11:05:41 1997
-- 
-- This is a Win32 specific driver for displaying animated images.

module ShowImageB
        ( disp    -- :: ImageB.Image -> IO ()
	, dispSave, dispEx

	-- all exported for use in FranPlugin
	, dispatch, worldToScreen, getTime
        ) where

import StaticTypes(
	Length,
	Vector2, vector2XY, vector2XYCoords,
	Point2,  point2XY,  point2XYCoords,
	Transform2, compose2, inverse2, uscale2, transform3x2,
	Transformable2((*%)),
	Translateable2(translate2),
	Scaleable2(scale2),
	Image,
	)
import Behavior( 
	Time, at 
	)
import PrimInteract( 
	PrimEV(fire),
	Input(..), inputVar,
	initInput
	)
import ImageB( 
	ImageB 
	)
import RenderImage(
	draw
	)

import qualified Win32
import Graphics hiding (draw)
import MVar( forkIO )
import IOExtensions( garbageCollect )
import IORef
import Monad( when, unless )
import Maybe( isJust, fromJust )

-- Outside entry point for real-time animations and no file writing.

disp :: (Time -> ImageB) -> IO ()

disp = dispEx Nothing makeRealTimeClock
 where
  makeRealTimeClock = do
    startMS <- Win32.timeGetTime
    return (getTime startMS)

-- Outside entry point for non-real-time animations written to a directory
-- full of BMP files.  Specify a directory name and a clock period in
-- seconds.  The directory will be created if it doesn't exist, and it
-- probably shouldn't because new contents would get mixed up with
-- existing contents.  Files with names wxyz.gif will get created in the
-- directory where "wxyz" is (1+t)*1000 if the frame corresponds to
-- animation time t.  (The "1+" is so that alphabetized file names will be
-- in time order.)  You can then use a tool like GIF Constructor Set by
-- Alchemy Mindworks (http://www.mindworkshop.com/alchemy/gifcon.html) to
-- make an animated GIFs.

dispSave :: String -> Double -> (Time -> ImageB) -> IO ()

dispSave directory clockPeriod imF = do
  Win32.createDirectory directory Nothing `catch` \ e ->
     putStrLn ("warning: " ++ directory ++ " already existed")
  dispEx (Just (directory ++ "/")) makeRegularClock imF
 where
  makeRegularClock = do
    lastTimeVar <- newRef 0
    return $ do
      tWas <- getRef lastTimeVar
      let tNow = tWas + clockPeriod
      setRef lastTimeVar tNow
      return tNow

-- extended version - lets you specify a directory to store bitmaps into
dispEx :: Maybe String -> IO (IO Time)
       -> (Time -> ImageB) -> IO ()
dispEx mbfile makeClock imF = runGraphics (do
  clock <- makeClock
  inputDevices <- getRef inputVar
  initInput inputDevices
  let (width, height) = fromVector2 defWindowSize
  w <- openWindowEx 
         "Fran Image Animation" 
         --(Just $ (1280-8-width, -20))  -- 1280x??? screen
	 (Just $ (1024-8-width, 0))    -- 1024x??? screen
         (Just $ (width, height))
         (drawBufferedPictureBC Win32.wHITENESS (RGB 0 0 0))
         (Just timerRate)
  let 
    drawImage :: Image -> Time -> IO ()
    drawImage im t = do
      -- garbageCollect -- forcing a GC makes things look smoother
      sz <- windowSize w
      let pic = worldToScreen sz *% im
      setPicture w $ draw pic
      when (isJust mbfile) $
        savePicture (fromJust mbfile ++ show (round ((1+t)*1000)) ++ ".bmp") 
	            sz 
		    (draw pic)

    drawLoop :: ImageB -> Time -> IO ()
    drawLoop imb t0 = do
      getTick w
      t <- clock

      -- This is a hack to make up for some bug which is
      -- causing Fran to ignore some resize events
      sz <- windowSize w
      fire (viewSz inputDevices) t (scale2 pixelSize *% toVector2 sz)

      let (im,imb') = imb `at` t
      drawImage im t `catch` (\ e -> putStrLn "Win32 drawing error")
      fire (fps inputDevices) t (if t==t0 then 0 else (1 / (t - t0)))
      drawLoop imb' t

  forkIO (drawLoop (imF 0) 0)
  let
    eventLoop :: IO ()
    eventLoop = do
      e <- getEvent w
      -- Don't look for input if we're writing files, since we'd be faking
      -- time.  Sorry for this hack.  It's really here so that dispatch
      -- doesn't invoke the clock, and so advance "time".  And if I put
      -- the "unless" around the main call to eventLoop below, dispEx
      -- exists without drawing anything.  (conal)
      unless (isJust mbfile) $
        dispatch w clock inputDevices e
      eventLoop
  eventLoop
  )
 where
  timerRate = round (1000 / targetFPS)

dispatch :: Window -> IO Time -> Input -> Event -> IO ()
dispatch w clock inputDevices e = do
      t <- clock
      case e of 
	Key { char, isDown } -> do
	    fire (key inputDevices) t (char, isDown)
	Button { pt, isLeft, isDown } -> do
	    p <- pointToPoint2 w pt
	    fire (button inputDevices) t (p,isLeft,isDown)
	MouseMove { pt } -> do
	    p <- pointToPoint2 w pt
	    fire (mousePos inputDevices) t p		
	-- Fran seems to be ignoring some of the "fire viewSz" events
	-- so we do some extra firings in the drawloop
	Resize -> do
	    sz <- windowSize w
            fire (viewSz inputDevices) t (scale2 pixelSize *% toVector2 sz)
        _ -> return ()

----------------------------------------------------------------
-- Scaled conversions
----------------------------------------------------------------

pointToPoint2  :: Window -> Point -> IO Point2
pointToVector2 :: Window -> Point -> IO Vector2
worldToScreen  :: Point -> Transform2
screenToWorld  :: Point -> Transform2

pointToPoint2 w pt = do
  sz <- windowSize w
  return (screenToWorld sz *% toPoint2 pt)

pointToVector2 w pt = do
  sz <- windowSize w
  return (screenToWorld sz *% toVector2 pt)

screenToWorld sz 
  | presentation
  = uscale2 m        `compose2` flipV `compose2` translate2 centre
  | otherwise
  = scale2 pixelSize `compose2` flipV `compose2` translate2 centre
 where
  centre = uscale2 (-1/2) *% toVector2 sz
  m = case sz of (w,h) -> (2.4 / fromInt (w `min` h))

worldToScreen = inverse2 . screenToWorld

-- flip vertically (to compensate for (0,0) being top-left corner)
flipV :: Transform2
flipV = transform3x2 1    0 0
                     0 (-1) 0

----------------------------------------------------------------
-- Raw coercions - ignore scalings
----------------------------------------------------------------

toPoint2    :: Point   -> Point2
toVector2   :: Point   -> Vector2
fromPoint2  :: Point2  -> Point
fromVector2 :: Vector2 -> Point

toPoint2    (x,y) = point2XY (fromInt x) (fromInt y)
toVector2   (x,y) = vector2XY (fromInt x) (fromInt y)
fromPoint2  z = case point2XYCoords  z of (x, y) -> (round x, round y)
fromVector2 z = case vector2XYCoords z of (x, y) -> (round x, round y)

----------------------------------------------------------------
-- Program parameters
----------------------------------------------------------------

-- Doing a (probably PowerPoint) presentation.  Make the window stay on
-- top, but in a corner.  Also, resizing the window scales the animation.
presentation = False

-- Size of each pixel on the screen.
-- If this corresponds to the values of
--   Image.bitmapPixelsPerLengthHorizontal
--   Image.bitmapPixelsPerLengthVertical  
-- they will cancel each other out.

pixelSize :: Vector2
pixelSize = vector2XY (1/100) (1/100)

-- Size of window on startup.
-- Big enough to show -1 to 1 in continuous units in X and Y, plus a bit.

defWindowSize :: Vector2
defWindowSize = inverse2 (scale2 pixelSize) *% vector2XY 2.1 2.1

-- Frames per second target
-- The maximum rate achievable on our system seems to be 20 fps -
-- no matter how trivial the image.
targetFPS = 20 :: Float

----------------------------------------------------------------
-- Win32 utilities
----------------------------------------------------------------

getTime :: Win32.MilliSeconds -> IO Time
getTime startMS = do
  ms <- Win32.timeGetTime
  return (fromIntegral (ms-startMS) / 1000.0)

----------------------------------------------------------------
