-----------------------------------------------------------------------------
-- Non-standard extensions to IO monad.
--
-- Binary file extensions:
--
--   readBinaryFile         : versions of readFile, writeFile, appendFile,
--   writeBinaryFile        : and openFile for use on binary files
--   appendBinaryFile       : (These don't do LF <-> CR-LF translation on
--   openBinaryFile         :  DOS/Windows systems.)
--
-- Miscellaneous extensions:
--
--   getCh                  : like getChar but doesn't echo to screen
--   fixIO                  : fixpoint operator for IO monad
--   argv                   : value returned by getArgv     
--   garbageCollect         : perform a garbage collection right now
-- 
-- Potentially dangerous extensions - may break referential transparency
-- or Haskell type system if misused.
--
--   unsafePerformIO        : subvert IO monad's linearity constraints
--   unsafeInterleaveIO     : (both are potentially dangerous)
--
-- None of these operations can be implemented in standard Haskell using the
-- standard Haskell prelude.
--
-- Suitable for use with Hugs 1.4.
-----------------------------------------------------------------------------

module IOExtensions(
	-- binary file operations:
	--   identical to text file operations on UNIX;
	--   don't do LF <-> CR-LF translation on DOS.
	readBinaryFile, writeBinaryFile, appendBinaryFile,
	openBinaryFile, 

        -- Miscellaneous extensions
        getCh,
	fixIO,
	argv,
	garbageCollect,

	-- semi-standard but dangerous sequential IO extensions
	unsafePerformIO,
	unsafeInterleaveIO,
	) where

import IO(IOMode, Handle, ioeGetErrorString) -- for binary file ops
import System( getArgs )

argv :: [String]
argv = unsafePerformIO getArgs

runAndShowError :: IO a -> IO a
runAndShowError m =
  m `catch` \err -> do 
      putChar '\n'
      putStr (ioeGetErrorString err)
      return undefined

unsafePerformIO :: IO a -> a
unsafePerformIO m = performIO (runAndShowError m)

unsafeInterleaveIO :: IO a -> IO a
unsafeInterleaveIO m = return (unsafePerformIO m)

fixIO :: (a -> IO a) -> IO a
-- OLD (wrong!): fixIO m = return a where a = unsafePerformIO (m a)
fixIO m = IO fixIO'
 where
  fixIO' fail succ = r
   where
    (IO ma) = m a
    r = ma fail succ
    a = case r of
	Hugs_Return a  -> a
	Hugs_Error err -> error "IOUtils.fixIO: thread exited with error"
	_              -> error "IOUtils.fixIO: thread exited with no result"

performIO :: IO a -> a
performIO (IO m) = 
  case m Hugs_Error Hugs_Return of
  Hugs_Return a  -> a
  Hugs_Error err -> error "IOUtils.performIO: thread exited with error"
  _              -> error "IOUtils.performIO: thread exited with no result"

primitive writeBinaryFile   	 :: FilePath -> String -> IO ()
primitive appendBinaryFile  	 :: FilePath -> String -> IO ()
primitive readBinaryFile    	 :: FilePath -> IO String
primitive openBinaryFile         :: FilePath -> IOMode -> IO Handle

primitive garbageCollect "primGC" :: IO ()
primitive getCh                   :: IO Char -- non-echoing getchar



