
Fran Bounded Events
~~~~~~~~~~~~~~~~~~~

Last modified Wed Nov 06 10:33:44 1996

Events generated using the code in External and PrimInteract suffer a
huge space leak: if you sample the event at 1 second intervals, the
event will grow into a thunk like this:

  Event (\ t -> if t < 1 then Occ 1 a1
        	else if t < 2 then Occ 2 a2
        	else if t < 3 then Occ 3 a3
        	else if t < 4 then Occ 4 a4
        	else ...
        )
        
Every time you evaluate it at a higher value of t, the thunk gets bigger.

In theory, there's no choice but to do this: the program might contain
an arbitrary time transform requiring input events from arbitrarily
far back in time.

In practice, there's no need to do this: programs never use time
transformations of more than a few seconds so we only need to buffer
input events received in the last few seconds.  Theoretically, this is
wrong, but we don't expect it to be a major problem in practice.

\begin{code}

module Bounded(
	module Bounded
	) where

import MutVar
import IOExtensions(unsafePerformIO)

import BaseTypes( Time )
import Event( Event(..), PossibleOcc(..) )

\end{code}


Event Buffers
~~~~~~~~~~~~~

Everything's built on this data structure: a list of time-value pairs
sorted wrt time.

\begin{code}

type Events a = [(Time,a)]  -- times are in increasing order

sample :: Time -> Events a -> Maybe (Time,a)
addEvent :: Time -> a -> Events a -> Events a

\end{code}

\begin{code}

sample t es = sample' Nothing t es

sample' :: Maybe (Time,a) -> Time -> Events a -> Maybe (Time,a)
sample' d t (e@(te,_):rest) 
  | t >= te
  = d
  | otherwise
  = sample' (Just e) t rest 
sample' d t [] = d

addEvent t a tas@(e@(te,_):rest)
  | t < te
  = (t,a) : tas
  | otherwise
  = e : addEvent t a rest
addEvent t a [] = [(t,a)]

\end{code}


Pruning Functions
~~~~~~~~~~~~~~~~~

Pruning functions are used to implement a bounding policy.  We provide
two different policies that you might use:

1) Keep the last n events.
2) Keep events for the last n seconds.

Both pruning functions have the property that

  prune . prune is hyperstrict

This is enough to prevent a frequently written but never read list
from building a large thunk consisting of stale data which will all
get deleted when we evaluate it.

\begin{code}

sizeBound :: Int  -> (Events a -> Events a)
timeBound :: Time -> (Events a -> Events a)

sizeBound n es = drop ((l-n) `max` 0) es
 where
  l = length es

timeBound dt [] = []
timeBound dt es = dropWhile (\ (te,_) -> te < t-dt) es
 where
  (t,_) = last es

\end{code}


Mutable Event Buffers
~~~~~~~~~~~~~~~~~~~~~

A mutable bounded event buffer is just like a PrimInteraction object
except that the mutable variable contains an explicit data structure
and that you provide an explicit pruning function (which might
restrict the buffer size or the time period covered by the buffer).

"fire" takes care to force the new event list to exploit the 
hyperstrictness (and thus avoid a space leak).

\begin{code}

data Show a => BoundedEv a = 
  BoundedEv {
    reset :: IO (),
    fire  :: Time -> a -> IO (),
    event :: Event a
    }

mkBoundedEv :: Show a => String -> (Events a -> Events a) -> IO (BoundedEv a) 
mkBoundedEv nm prune = do
  buffer <- newVar (error errmsg1)
  return $ BoundedEv 
    { reset = do
        writeVar buffer []
    , fire = \ t a -> do
	es <- readVar buffer
	print es
	strict (writeVar buffer) (prune (addEvent t a es))
    , event = let sample' t = unsafePerformIO $ do
		    es <- readVar buffer
		    print (t, sample t es)
		    case sample t es of
		    Just (te,a) -> return $ Occ te a
		    Nothing     -> return $ NonOcc event'
                  event' = Event sample'
              in event'
    }
 where
  errmsg1 = "Event Buffer " ++ nm ++ " used without being reset."

\end{code}

